from tabulate import tabulate
import sys
import os
import click

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)
from utils import PyOrm, Completer, getPackageVersion, WELCOME_MSG, EXIT_MSG
from Fs import readCsvFile
from Validation import (
    validateParamsByCustomFunc,
    validateCollectionParameter,
    validateIndexParameter,
    validateSearchParams,
    validateQueryParams,
    validateCalcParams,
)
from Types import ParameterException, ConnectException
from Types import MetricTypes, IndexTypesMap, IndexTypes

pass_context = click.make_pass_decorator(PyOrm, ensure=True)


@click.group(no_args_is_help=False,
             add_help_option=False,
             invoke_without_command=True)
@click.pass_context
def cli(ctx):
    """Milvus_CLI"""
    ctx.obj = PyOrm()


def print_help_msg(command):
    with click.Context(command) as ctx:
        click.echo(command.get_help(ctx))


@cli.command()
def help():
    """Show help messages."""
    click.echo(print_help_msg(cli))


@cli.command(no_args_is_help=False)
@click.option(
    "-a",
    "--alias",
    "alias",
    help="[Optional] - Milvus link alias name, default is `default`.",
    default="default",
    type=str,
)
@click.option(
    "-h",
    "--host",
    "host",
    help="[Optional] - Host name, default is `127.0.0.1`.",
    default="127.0.0.1",
    type=str,
)
@click.option(
    "-p",
    "--port",
    "port",
    help="[Optional] - Port, default is `19530`.",
    default=19530,
    type=int,
)
@click.option(
    "-s",
    "--secure",
    "secure",
    help="[Optional] - Secure, default is `False`.",
    default=False,
    type=bool,
)
@click.option(
    "-u",
    "--username",
    "username",
    help="[Optional] - Username , default is `None`.",
    default=None,
    type=str,
)
@click.option(
    "-pwd",
    "--password",
    "password",
    help="[Optional] - Password , default is `None`.",
    default=None,
    type=str,
)
@click.option(
    "-D",
    "--disconnect",
    "disconnect",
    help=
    "[Optional, Flag] - Disconnect from a Milvus server by alias, default is `default`.",
    default=False,
    is_flag=True,
)
@click.pass_obj
def connect(obj, alias, host, port, secure, username, password, disconnect):
    """
    Connect to Milvus.

    Example:

        milvus_cli > connect -h 127.0.0.1 -p 19530 -a default
    """
    try:
        obj.connect(alias, host, port, disconnect, secure, username, password)
    except Exception as e:
        click.echo(message=e, err=True)
    else:
        if disconnect:
            click.echo("Disconnected.")
        else:
            click.echo("Connect Milvus successfully.")
            click.echo(obj.showConnection(alias))


@cli.command()
def version():
    """Get Milvus_CLI version."""
    click.echo(f"Milvus_CLI v{getPackageVersion()}")


@cli.command()
def clear():
    """Clear screen."""
    click.clear()


@cli.group(no_args_is_help=False)
@click.pass_obj
def show(obj):
    """Show connection, loading_progress and index_progress."""
    pass


@show.command()
@click.option(
    "-a",
    "--all",
    "showAll",
    help="[Optional, Flag] - Show all connections.",
    default=False,
    is_flag=True,
)
@click.pass_obj
def connection(obj, showAll):
    """Show current/all connection details"""
    try:
        (not showAll) and obj.checkConnection()
    except Exception as e:
        click.echo("No connections.")
    else:
        click.echo(obj.showConnection(showAll=showAll))


@show.command("loading_progress")
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection is loading")
@click.option(
    "-p",
    "--partition",
    "partition",
    help="[Optional, Multiple] - The names of partitions are loading",
    default=None,
    multiple=True,
)
@click.pass_obj
def loadingProgress(obj, collection, partition):
    """Show #loaded entities vs #total entities."""
    try:
        obj.checkConnection()
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        result = obj.showCollectionLoadingProgress(collection, partition)
    except Exception as e:
        click.echo(message=e, err=True)
    else:
        click.echo(
            tabulate(
                [[
                    result.get("num_loaded_entities"),
                    result.get("num_total_entities")
                ]],
                headers=["num_loaded_entities", "num_total_entities"],
                tablefmt="pretty",
            ))


@show.command("index_progress")
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection is loading")
# ! TODO: To be removed
@click.option("-i",
              "--index",
              "index",
              help="[Optional] - Index name.",
              default="")
@click.pass_obj
def indexProgress(obj, collection, index):
    """Show # indexed entities vs. # total entities."""
    try:
        obj.checkConnection()
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        result = obj.showIndexBuildingProgress(collection, index)
    except Exception as e:
        click.echo(message=e, err=True)
    else:
        click.echo(
            tabulate(
                [[result.get("indexed_rows"),
                  result.get("total_rows")]],
                headers=["indexed_rows", "total_rows"],
                tablefmt="pretty",
            ))


@show.command("query_segment")
@click.option(
    "-c",
    "--collection-name",
    "collection",
    help="A string representing the collection to get segments info.",
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. When timeout is not set, client waits until server response or error occur.",
    default=None,
    type=float,
)
@click.pass_obj
def querySegmentInfo(obj, collection, timeout):
    """Return segments information from query nodes."""
    click.echo(
        obj.getQuerySegmentInfo(collection, timeout, prettierFormat=True))


@show.command("compaction_state")
@click.option(
    "-c",
    "--collection-name",
    "collection",
    help="Collection name.",
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. When timeout is not set, client waits until server response or error occur.",
    default=None,
    type=float,
)
@click.pass_obj
def compactionStateInfo(obj, collection, timeout):
    """Return the current collection's compaction state."""
    click.echo(obj.getCompactCollectionState(collection, timeout))


@show.command("compaction_plans")
@click.option(
    "-c",
    "--collection-name",
    "collection",
    help="Collection name.",
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. When timeout is not set, client waits until server response or error occur.",
    default=None,
    type=float,
)
@click.pass_obj
def compactionStateInfo(obj, collection, timeout):
    """Return the current collection's compaction state."""
    click.echo(obj.getCompactCollectionPlans(collection, timeout))


@cli.command()
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection to load.")
@click.option(
    "-p",
    "--partition",
    "partition",
    help="[Optional, Multiple] - The name of partition to load.",
    default=[],
    multiple=True,
)
@click.pass_obj
def load(obj, collection, partition):
    """Load specified collection/partitions from disk to memory."""
    try:
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        for partitionName in partition:
            validateParamsByCustomFunc(
                obj.getTargetPartition,
                "Partition Name Error!",
                collection,
                partitionName,
            )
        if partition:
            result = obj.loadPartitions(collection, partition)
        else:
            result = obj.loadCollection(collection)
    except Exception as e:
        click.echo(message=e, err=True)
    else:
        if partition:
            click.echo(
                f"""Load {collection}'s partitions {partition} successfully""")
        else:
            click.echo(f"""Load Collection {collection} successfully""")
        click.echo(result)


@cli.command()
@click.option(
    "-c",
    "--collection-name",
    "collection",
    help="The name of collection to be released.",
)
@click.option(
    "-p",
    "--partition",
    "partition",
    help="[Optional, Multiple] - The name of partition to released.",
    default=[],
    multiple=True,
)
@click.pass_obj
def release(obj, collection, partition):
    """Release specified collection/partitions from memory."""
    try:
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        for partitionName in partition:
            validateParamsByCustomFunc(
                obj.getTargetPartition,
                "Partition Name Error!",
                collection,
                partitionName,
            )
        if partition:
            result = obj.releasePartitions(collection, partition)
        else:
            result = obj.releaseCollection(collection)
    except Exception as e:
        click.echo(message=e, err=True)
    else:
        if partition:
            click.echo(
                f"""Release {collection}'s partitions {partition} successfully"""
            )
        else:
            click.echo(f"""Release Collection {collection} successfully""")
        click.echo(result)


@cli.group("list", no_args_is_help=False)
@click.pass_obj
def listDetails(obj):
    """List collections, partitions and indexes."""
    pass


# @listDetails.command("alias")
# @click.option("-c", "--collection-name", "collection", help="The name of collection.")
# @click.option(
#     "--timeout",
#     "-t",
#     "timeout",
#     help="[Optional] - An optional duration of time in seconds to allow for the RPC. When timeout is not set, client waits until server response or error occur.",
#     default=None,
#     type=float,
# )
# @click.pass_obj
# def listCollectionAlias(obj, collection, timeout):
#     """List all alias of the collection."""
#     try:
#         obj.checkConnection()
#         click.echo(obj.listCollectionAlias(collection, timeout))
#     except Exception as e:
#         click.echo(message=e, err=True)


@listDetails.command()
@click.option(
    "--timeout",
    "-t",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. When timeout is not set, client waits until server response or error occur.",
    default=None,
    type=float,
)
@click.pass_obj
def collections(obj, timeout):
    """List all collections."""
    try:
        obj.checkConnection()
        click.echo(obj.listCollections(timeout))
    except Exception as e:
        click.echo(message=e, err=True)


@listDetails.command()
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection.")
@click.pass_obj
def partitions(obj, collection):
    """List all partitions of the specified collection."""
    try:
        obj.checkConnection()
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        click.echo(obj.listPartitions(collection))
    except Exception as e:
        click.echo(message=e, err=True)


@listDetails.command()
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection.")
@click.pass_obj
def indexes(obj, collection):
    """List all indexes of the specified collection."""
    try:
        obj.checkConnection()
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collection)
        click.echo(obj.listIndexes(collection))
    except Exception as e:
        click.echo(message=e, err=True)


@listDetails.command()
@click.pass_obj
def users(obj):
    """List all users in Milvus"""
    try:
        obj.checkConnection()
        click.echo(obj.listCredUsers())
    except Exception as e:
        click.echo(message=e, err=True)


@cli.group("describe", no_args_is_help=False)
@click.pass_obj
def describeDetails(obj):
    """Describe collection, partition and index."""
    pass


@describeDetails.command("collection")
@click.option("-c",
              "--collection-name",
              "collection",
              help="The name of collection.")
@click.pass_obj
def describeCollection(obj, collection):
    """
    Describe collection.

    Example:

        milvus_cli > describe collection -c test_collection_insert
    """
    try:
        obj.checkConnection()
        click.echo(obj.getCollectionDetails(collection))
    except Exception as e:
        click.echo(message=e, err=True)


@describeDetails.command("partition")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="The name of collection.")
@click.option(
    "-p",
    "--partition",
    "partition",
    help='[Optional] - The name of partition, default is "_default".',
    default="_default",
)
@click.pass_obj
def describePartition(obj, collectionName, partition):
    """
    Describe partition.

    Example:

        milvus_cli > describe partition -c test_collection_insert -p _default
    """
    try:
        obj.checkConnection()
        collection = obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error when getting collection by name!\n{str(e)}")
    else:
        click.echo(obj.getPartitionDetails(collection, partition))


@describeDetails.command("index")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="The name of collection.")
@click.option("-in", "--index-name", "indexName", help="The name of index.")
@click.pass_obj
def describeIndex(obj, collectionName, indexName):
    """
    Describe index.

    Example:

        milvus_cli > describe index -c car -in vectorIndex
    """
    try:
        obj.checkConnection()
        collection = obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error when getting collection by name!\n{str(e)}")
    else:
        click.echo(obj.getIndexDetails(collection, indexName))


@cli.group("create", no_args_is_help=False)
@click.pass_obj
def createDetails(obj):
    """Create collection, partition and index."""
    pass


@createDetails.command("alias")
@click.option(
    "-c",
    "--collection-name",
    "collectionName",
    help="Collection name to be specified alias.",
    type=str,
)
@click.option(
    "-a",
    "--alias-name",
    "aliasNames",
    help="[Multiple] - The alias of the collection.",
    type=str,
    multiple=True,
)
@click.option(
    "-A",
    "--alter",
    "alter",
    help="[Optional, Flag] - Change an existing alias to current collection.",
    default=False,
    is_flag=True,
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def createAlias(obj, collectionName, aliasNames, alter, timeout):
    """
    Specify alias for a collection.
    Alias cannot be duplicated, you can't assign same alias to different collections.
    But you can specify multiple aliases for a collection, for example:

    create alias -c car -a carAlias1 -a carAlias2

    You can also change alias of a collection to another collection.
    If the alias doesn't exist, it will return error.
    Use "-A" option to change alias owner collection, for example:

    create alias -c car2 -A -a carAlias1 -a carAlias2
    """
    try:
        obj.checkConnection()
        if alter:
            result = obj.alterCollectionAliasList(collectionName, aliasNames,
                                                  timeout)
        else:
            result = obj.createCollectionAliasList(collectionName, aliasNames,
                                                   timeout)
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        if len(result) == len(aliasNames):
            click.echo(
                f"""{len(result)} alias {"altered" if alter else "created"} successfully."""
            )


@createDetails.command("collection")
@click.option(
    "-c",
    "--collection-name",
    "collectionName",
    help="Collection name to specify alias.",
    type=str,
)
@click.option("-p",
              "--schema-primary-field",
              "primaryField",
              help="Primary field name.")
@click.option(
    "-a",
    "--schema-auto-id",
    "autoId",
    help="[Optional, Flag] - Enable auto id.",
    default=False,
    is_flag=True,
)
@click.option(
    "-d",
    "--schema-description",
    "description",
    help="[Optional] - Description details.",
    default="",
)
@click.option(
    "-f",
    "--schema-field",
    "fields",
    help=
    '[Multiple] - FieldSchema. Usage is "<Name>:<DataType>:<Dim(if vector) or Description>"',
    default=None,
    multiple=True,
)
@click.pass_obj
def createCollection(obj, collectionName, primaryField, autoId, description,
                     fields):
    """
    Create collection.

    Example:

      create collection -c car -f id:INT64:primary_field -f vector:FLOAT_VECTOR:128 -f color:INT64:color -f brand:INT64:brand -p id -a -d 'car_collection'
    """
    try:
        obj.checkConnection()
        validateCollectionParameter(collectionName, primaryField, fields)
    except ParameterException as pe:
        click.echo("Error!\n{}".format(str(pe)))
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        click.echo(
            obj.createCollection(collectionName, primaryField, autoId,
                                 description, fields))
        click.echo("Create collection successfully!")


@createDetails.command("partition")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="Collection name.")
@click.option("-p", "--partition", "partition", help="The name of partition.")
@click.option(
    "-d",
    "--description",
    "description",
    help="[Optional] - Partition description.",
    default="",
)
@click.pass_obj
def createPartition(obj, collectionName, partition, description):
    """
    Create partition.

    Example:

        milvus_cli > create partition -c car -p new_partition -d test_add_partition
    """
    try:
        obj.checkConnection()
        obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error occurred when get collection by name!\n{str(e)}")
    else:
        click.echo(obj.createPartition(collectionName, description, partition))
        click.echo("Create partition successfully!")


@createDetails.command("index")
@click.pass_obj
def createIndex(obj):
    """
    Create index.

    Example:

        milvus_cli > create index

        Collection name (car, car2): car2

        The name of the field to create an index for (vector): vector

        Index type (FLAT, IVF_FLAT, IVF_SQ8, IVF_PQ, RNSG, HNSW, ANNOY): IVF_FLAT

        Index metric type (L2, IP, HAMMING, TANIMOTO): L2

        Index params nlist: 2

        Timeout []:
    """
    try:
        obj.checkConnection()
        collectionName = click.prompt("Collection name",
                                      type=click.Choice(
                                          obj._list_collection_names()))
        fieldName = click.prompt(
            "The name of the field to create an index for",
            type=click.Choice(
                obj._list_field_names(collectionName, showVectorOnly=True)),
        )
        indexName = click.prompt("Index name")

        indexType = click.prompt("Index type", type=click.Choice(IndexTypes))
        metricType = click.prompt("Index metric type",
                                  type=click.Choice(MetricTypes))
        index_building_parameters = IndexTypesMap[indexType][
            "index_building_parameters"]
        params = []
        for param in index_building_parameters:
            tmpParam = click.prompt(f"Index params {param}")
            params.append(f"{param}:{tmpParam}")
        inputTimeout = click.prompt("Timeout", default="")
        timeout = inputTimeout if inputTimeout else None
        validateIndexParameter(indexType, metricType, params)
    except ParameterException as pe:
        click.echo("Error!\n{}".format(str(pe)))
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        click.echo(
            obj.createIndex(collectionName, fieldName, indexName, indexType,
                            metricType, params, timeout))
        click.echo("Create index successfully!")


@createDetails.command("user")
@click.option("-u",
              "--username",
              "username",
              help="The username of milvus user.")
@click.option("-p",
              "--password",
              "password",
              help="The pawssord of milvus user.")
@click.pass_obj
def createUser(obj, username, password):
    """
    Create user.

    Example:

        milvus_cli > create user -u zilliz -p zilliz
    """
    try:
        obj.checkConnection()
        click.echo(obj.createCredUser(username, password))
        click.echo("Create user successfully")
    except Exception as e:
        click.echo(message=e, err=True)


@cli.group("delete", no_args_is_help=False)
@click.pass_obj
def deleteObject(obj):
    """Delete specified collection, partition and index."""
    pass


@deleteObject.command("alias")
@click.option("-a",
              "--alias-name",
              "aliasName",
              help="The alias of the collection.",
              type=str)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def deleteAlias(obj, aliasName, timeout):
    """
    Delete an alias.
    """
    click.echo(
        "Warning!\nYou are trying to delete an alias. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        obj.dropCollectionAlias(aliasName, timeout)
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        click.echo(f"Drop alias '{aliasName}' successfully.")


@deleteObject.command("collection")
@click.option(
    "-c",
    "--collection-name",
    "collectionName",
    help="The name of collection to be deleted.",
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def deleteCollection(obj, collectionName, timeout):
    """
    Drops the collection together with its index files.

    Example:

        milvus_cli > delete collection -c car
    """
    click.echo(
        "Warning!\nYou are trying to delete the collection with data. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error occurred when get collection by name!\n{str(e)}")
    else:
        result = obj.dropCollection(collectionName, timeout)
        click.echo("Drop collection successfully!"
                   ) if not result else click.echo("Drop collection failed!")


@deleteObject.command("partition")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="Collection name")
@click.option("-p", "--partition", "partition", help="The name of partition.")
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def deletePartition(obj, collectionName, partition, timeout):
    """
    Drop the partition and its corresponding index files.

    Example:

        milvus_cli > delete partition -c car -p new_partition
    """
    click.echo(
        "Warning!\nYou are trying to delete the partition with data. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error occurred when get collection by name!\n{str(e)}")
    else:
        result = obj.dropPartition(collectionName, partition, timeout)
        click.echo("Drop partition successfully!"
                   ) if not result else click.echo("Drop partition failed!")


@deleteObject.command("index")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="Collection name")
@click.option("-in", "--index-name", "Index name", help="Index name")
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def deleteIndex(obj, collectionName, indexName, timeout):
    """
    Drop index and its corresponding index files.

    Example:

        milvus_cli > delete index -c car
    """
    click.echo(
        "Warning!\nYou are trying to delete the index of collection. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error occurred when get collection by name!\n{str(e)}")
    else:
        result = obj.dropIndex(collectionName, indexName, timeout)
        click.echo("Drop index successfully!") if not result else click.echo(
            "Drop index failed!")


@deleteObject.command("user")
@click.option("-u",
              "--username",
              "username",
              help="The username of milvus user")
@click.pass_obj
def deleteUser(obj, username):
    """
    Drop user in milvus by username

    Example:

        milvus_cli > delete user -u zilliz
    """
    click.echo(
        "Warning!\nYou are trying to delete the user in milvus. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        result = obj.deleteCredUser(username)
        click.echo("Drop user successfully!")
        click.echo(result)
    except Exception as e:
        click.echo(message=e, err=True)


@deleteObject.command("entities")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="Collection name.")
@click.option(
    "-p",
    "--partition",
    "partitionName",
    help="[Optional] - Name of partitions that contain entities.",
    default="",
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def deleteEntities(obj, collectionName, partitionName, timeout):
    """
    Delete entities with an expression condition. And return results to show which primary key is deleted successfully.

    Example:

        milvus_cli > delete entities -c car

        The expression to specify entities to be deleted, such as "film_id in [ 0, 1 ]": film_id in [ 0, 1 ]

        You are trying to delete the entities of collection. This action cannot be undone!

        Do you want to continue? [y/N]: y
    """
    expr = click.prompt(
        '''The expression to specify entities to be deleted, such as "film_id in [ 0, 1 ]"'''
    )
    click.echo(
        "You are trying to delete the entities of collection. This action cannot be undone!\n"
    )
    if not click.confirm("Do you want to continue?"):
        return
    try:
        obj.checkConnection()
        obj.getTargetCollection(collectionName)
    except Exception as e:
        click.echo(f"Error occurred when get collection by name!\n{str(e)}")
    else:
        partitionValue = partitionName if partitionName else None
        timeoutValue = timeout if timeout else None
        result = obj.deleteEntities(expr, collectionName, partitionValue,
                                    timeoutValue)
        click.echo(result)


@cli.command()
@click.pass_obj
def search(obj):
    """
    Conducts a vector similarity search with an optional boolean expression as filter.

    Example-1(import a CSV file):

        Collection name (car, test_collection): car

        The vectors of search data (the length of data is number of query (nq),
        the dim of every vector in data must be equal to vector field’s of
        collection. You can also import a CSV file without headers): examples/import_csv/search_vectors.csv

        The vector field used to search of collection (vector): vector

        Metric type: L2

        Search parameter nprobe's value: 10

        The max number of returned record, also known as topk: 2

        The boolean expression used to filter attribute []: id > 0

        The names of partitions to search (split by "," if multiple) ['_default'] []: _default

    Example-2(collection has index):

        Collection name (car, test_collection): car

        \b
        The vectors of search data (the length of data is number of query (nq),
        the dim of every vector in data must be equal to vector field’s of
        collection. You can also import a CSV file without headers):
            [[0.71, 0.76, 0.17, 0.13, 0.42, 0.07, 0.15, 0.67, 0.58, 0.02, 0.39,
            0.47, 0.58, 0.88, 0.73, 0.31, 0.23, 0.57, 0.33, 0.2, 0.03, 0.43,
            0.78, 0.49, 0.17, 0.56, 0.76, 0.54, 0.45, 0.46, 0.05, 0.1, 0.43,
            0.63, 0.29, 0.44, 0.65, 0.01, 0.35, 0.46, 0.66, 0.7, 0.88, 0.07,
            0.49, 0.92, 0.57, 0.5, 0.16, 0.77, 0.98, 0.1, 0.44, 0.88, 0.82,
            0.16, 0.67, 0.63, 0.57, 0.55, 0.95, 0.13, 0.64, 0.43, 0.71, 0.81,
            0.43, 0.65, 0.76, 0.7, 0.05, 0.24, 0.03, 0.9, 0.46, 0.28, 0.92,
            0.25, 0.97, 0.79, 0.73, 0.97, 0.49, 0.28, 0.64, 0.19, 0.23, 0.51,
            0.09, 0.1, 0.53, 0.03, 0.23, 0.94, 0.87, 0.14, 0.42, 0.82, 0.91,
            0.11, 0.91, 0.37, 0.26, 0.6, 0.89, 0.6, 0.32, 0.11, 0.98, 0.67,
            0.12, 0.66, 0.47, 0.02, 0.15, 0.6, 0.64, 0.57, 0.14, 0.81, 0.75,
            0.11, 0.49, 0.78, 0.16, 0.63, 0.57, 0.18]]

        The vector field used to search of collection (vector): vector

        Metric type: L2

        Search parameter nprobe's value: 10

        The specified number of decimal places of returned distance [-1]: 5

        The max number of returned record, also known as topk: 2

        The boolean expression used to filter attribute []: id > 0

        The names of partitions to search (split by "," if multiple) ['_default'] []: _default

        timeout []:

    Example-3(collection has no index):

        Collection name (car, car2): car

        The vectors of search data( the length of data is number of query (nq),
        the dim of every vector in data must be equal to vector field’s of
        collection. You can also import a CSV file without headers): examples/import_csv/search_vectors.csv

        The vector field used to search of collection (vector): vector

        The specified number of decimal places of returned distance [-1]: 5

        The max number of returned record, also known as topk: 2

        The boolean expression used to filter attribute []:

        The names of partitions to search (split by "," if multiple) ['_default'] []:

        Timeout []:
    """
    collectionName = click.prompt("Collection name",
                                  type=click.Choice(
                                      obj._list_collection_names()))
    data = click.prompt(
        "The vectors of search data (the length of data is number of query (nq), the dim of every vector in data must be equal to vector field’s of collection. You can also import a CSV file without headers)"
    )
    annsField = click.prompt(
        "The vector field used to search of collection",
        type=click.Choice(
            obj._list_field_names(collectionName, showVectorOnly=True)),
    )
    indexDetails = obj._list_index(collectionName)
    hasIndex = not not indexDetails
    if indexDetails:
        index_type = indexDetails["index_type"]
        search_parameters = IndexTypesMap[index_type]["search_parameters"]
        metric_type = indexDetails["metric_type"]
        click.echo(f"Metric type: {metric_type}")
        metricType = metric_type
        params = []
        for parameter in search_parameters:
            paramInput = click.prompt(f"Search parameter {parameter}'s value")
            params += [f"{parameter}:{paramInput}"]
    else:
        metricType = ""
        params = []
    roundDecimal = click.prompt(
        "The specified number of decimal places of returned distance",
        default=-1,
        type=int,
    )
    limit = click.prompt(
        "The max number of returned record, also known as topk",
        default=None,
        type=int)
    expr = click.prompt("The boolean expression used to filter attribute",
                        default="")
    partitionNames = click.prompt(
        f'The names of partitions to search (split by "," if multiple) {obj._list_partition_names(collectionName)}',
        default="",
    )
    timeout = click.prompt("Timeout", default="")
    guarantee_timestamp = click.prompt(
        "Guarantee Timestamp(It instructs Milvus to see all operations performed before a provided timestamp. If no such timestamp is provided, then Milvus will search all operations performed to date)",
        default=0,
        type=int,
    )
    travel_timestamp = click.prompt(
        "Travel Timestamp(Specify a timestamp in a search to get results based on a data view)",
        default=0,
        type=int,
    )
    export, exportPath = False, ""
    # if click.confirm('Would you like to export results as a CSV file?'):
    #     export = True
    #     exportPath = click.prompt('Directory path to csv file')
    # export = click.prompt('Would you like to export results as a CSV file?', default='n', type=click.Choice(['Y', 'n']))
    # if export:
    #     exportPath = click.prompt('Directory path to csv file')
    try:
        searchParameters = validateSearchParams(
            data,
            annsField,
            metricType,
            params,
            limit,
            expr,
            partitionNames,
            timeout,
            roundDecimal,
            hasIndex=hasIndex,
            guarantee_timestamp=guarantee_timestamp,
            travel_timestamp=travel_timestamp,
        )
        obj.checkConnection()
    except ParameterException as pe:
        click.echo("Error!\n{}".format(str(pe)))
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        if export:
            results = obj.search(collectionName,
                                 searchParameters,
                                 prettierFormat=False)
        else:
            results = obj.search(collectionName, searchParameters)
            click.echo(f"Search results:\n")
            for idx, item in enumerate(results):
                click.echo(f"No.{idx+1}:\n{item}\n")
            # click.echo(obj.search(collectionName, searchParameters))


@cli.command()
@click.pass_obj
def query(obj):
    """
    Query with a set of criteria, and results in a list of records that match the query exactly.

    Example 1:

        milvus_cli > query

        Collection name: car

        The query expression: id in [ 428960801420883491, 428960801420883492, 428960801420883493 ]

        Name of partitions that contain entities(split by "," if multiple) []: default

        A list of fields to return(split by "," if multiple) []: color, brand

        timeout []:

        Guarantee timestamp. This instructs Milvus to see all operations performed before a provided timestamp. If no such timestamp is provided, then Milvus will search all operations performed to date. [0]:

        Graceful time. Only used in bounded consistency level. If graceful_time is set, PyMilvus will use current timestamp minus the graceful_time as the guarantee_timestamp. This option is 5s by default if not set. [5]:

        Travel timestamp. Users can specify a timestamp in a search to get results based on a data view at a specified point in time. [0]: 428960801420883491

    Example 2:

        milvus_cli > query

        Collection name: car

        The query expression: id > 428960801420883491

        Name of partitions that contain entities(split by "," if multiple) []: default

        A list of fields to return(split by "," if multiple) []: id, color, brand

        timeout []:

        Guarantee timestamp. This instructs Milvus to see all operations performed before a provided timestamp. If no such timestamp is provided, then Milvus will search all operations performed to date. [0]:

        Graceful time. Only used in bounded consistency level. If graceful_time is set, PyMilvus will use current timestamp minus the graceful_time as the guarantee_timestamp. This option is 5s by default if not set. [5]:

        Travel timestamp. Users can specify a timestamp in a search to get results based on a data view at a specified point in time. [0]: 428960801420883491
    """
    collectionName = click.prompt("Collection name",
                                  type=click.Choice(
                                      obj._list_collection_names()))
    expr = click.prompt("The query expression")
    partitionNames = click.prompt(
        f'The names of partitions to search (split by "," if multiple) {obj._list_partition_names(collectionName)}',
        default="",
    )
    outputFields = click.prompt(
        f'Fields to return(split by "," if multiple) {obj._list_field_names(collectionName)}',
        default="",
    )
    timeout = click.prompt("timeout", default="")
    guarantee_timestamp = click.prompt(
        "Guarantee timestamp. This instructs Milvus to see all operations performed before a provided timestamp. If no such timestamp is provided, then Milvus will search all operations performed to date.",
        default=0,
        type=int,
    )
    graceful_time = click.prompt(
        "Graceful time. Only used in bounded consistency level. If graceful_time is set, PyMilvus will use current timestamp minus the graceful_time as the guarantee_timestamp. This option is 5s by default if not set.",
        default=5,
        type=int,
    )
    travel_timestamp = click.prompt(
        "Travel timestamp. Users can specify a timestamp in a search to get results based on a data view at a specified point in time.",
        default=0,
        type=int,
    )
    try:
        queryParameters = validateQueryParams(
            expr,
            partitionNames,
            outputFields,
            timeout,
            guarantee_timestamp,
            graceful_time,
            travel_timestamp,
        )
        obj.checkConnection()
    except ParameterException as pe:
        click.echo("Error!\n{}".format(str(pe)))
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        click.echo(obj.query(collectionName, queryParameters))


@cli.command("import")
@click.option(
    "-c",
    "--collection-name",
    "collectionName",
    help="The name of collection to be imported.",
)
@click.option(
    "-p",
    "--partition",
    "partitionName",
    help=
    "[Optional] - The partition name which the data will be inserted to, if partition name is not passed, then the data will be inserted to “_default” partition.",
    default=None,
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.argument("path")
@click.pass_obj
def importData(obj, collectionName, partitionName, timeout, path):
    """
    Import data from csv file(local or remote) with headers and insert into target collection.

    Example-1:

        milvus_cli > import -c car 'examples/import_csv/vectors.csv'

        Reading file from local path.

        Reading csv file...  [####################################]  100%

        Column names are ['vector', 'color', 'brand']

        Processed 50001 lines.

        Inserting ...

        Insert successfully.

        \b
    --------------------------  ------------------
    Total insert entities:                   50000
    Total collection entities:              150000
    Milvus timestamp:           428849214449254403
    --------------------------  ------------------

    Example-2:

        milvus_cli > import -c car 'https://raw.githubusercontent.com/zilliztech/milvus_cli/main/examples/import_csv/vectors.csv'

        Reading file from remote URL.

        Reading csv file...  [####################################]  100%

        Column names are ['vector', 'color', 'brand']

        Processed 50001 lines.

        Inserting ...

        Insert successfully.

        \b
    --------------------------  ------------------
    Total insert entities:                   50000
    Total collection entities:              150000
    Milvus timestamp:           428849214449254403
    --------------------------  ------------------
    """
    try:
        obj.checkConnection()
        validateParamsByCustomFunc(obj.getTargetCollection,
                                   "Collection Name Error!", collectionName)
        result = readCsvFile(path.replace('"', "").replace("'", ""))
        data = result["data"]
        result = obj.importData(collectionName, data, partitionName, timeout)
    except Exception as e:
        click.echo("Error!\n{}".format(str(e)))
    else:
        click.echo(f"\nInserted successfully.\n")
        click.echo(result)


@cli.group("calc", no_args_is_help=False)
@click.pass_obj
def calcUtils(obj):
    """Calculate distance, mkts_from_hybridts, mkts_from_unixtime and hybridts_to_unixtime."""
    pass


@calcUtils.command("distance")
@click.pass_obj
def calcDistance(obj):
    """
    Calculate distance between two vector arrays.

    Example:

        milvus_cli > calc distance

        Import left operator vectors from existing collection? [y/N]: n

        The vector's type (float_vectors, bin_vectors): float_vectors

        \b
        Left vectors:
            [[0.083, 0.992, 0.931, 0.433, 0.93, 0.706, 0.668, 0.481, 0.255, 0.088,
            0.121, 0.701, 0.935, 0.142, 0.012, 0.197, 0.066, 0.864, 0.263, 0.732,
            0.445, 0.672, 0.184, 0.675, 0.361, 0.115, 0.396, 0.206, 0.084, 0.274,
            0.523, 0.958, 0.071, 0.646, 0.864, 0.434, 0.212, 0.5, 0.319, 0.608,
            0.356, 0.745, 0.672, 0.488, 0.221, 0.485, 0.193, 0.557, 0.546, 0.626,
            0.593, 0.526, 0.404, 0.795, 0.076, 0.156, 0.231, 0.1, 0.18, 0.796,
            0.716, 0.752, 0.816, 0.363], [0.284, 0.135, 0.172, 0.198, 0.752, 0.174,
            0.314, 0.18, 0.672, 0.727, 0.062, 0.611, 0.921, 0.851, 0.238, 0.648,
            0.794, 0.177, 0.639, 0.339, 0.402, 0.977, 0.887, 0.528, 0.768, 0.16,
            0.698, 0.016, 0.906, 0.261, 0.902, 0.93, 0.547, 0.146, 0.65, 0.072,
            0.876, 0.645, 0.303, 0.922, 0.807, 0.093, 0.063, 0.344, 0.667, 0.81,
            0.662, 0.147, 0.242, 0.641, 0.903, 0.714, 0.637, 0.365, 0.512, 0.267,
            0.577, 0.809, 0.698, 0.62, 0.768, 0.402, 0.922, 0.592]]

        Import right operator vectors from existing collection? [y/N]: n

        The vector's type (float_vectors, bin_vectors): float_vectors

        \b
        Right vectors:
            [[0.518, 0.034, 0.786, 0.251, 0.04, 0.247, 0.55, 0.595, 0.638, 0.957,
            0.303, 0.023, 0.007, 0.712, 0.841, 0.648, 0.807, 0.429, 0.402, 0.904,
            0.002, 0.882, 0.69, 0.268, 0.732, 0.511, 0.942, 0.202, 0.749, 0.234,
            0.666, 0.517, 0.787, 0.399, 0.565, 0.457, 0.57, 0.937, 0.712, 0.981,
            0.928, 0.678, 0.154, 0.775, 0.754, 0.532, 0.074, 0.493, 0.288, 0.229,
            0.9, 0.657, 0.936, 0.184, 0.478, 0.587, 0.592, 0.84, 0.793, 0.985,
            0.826, 0.595, 0.947, 0.175], [0.704, 0.02, 0.937, 0.249, 0.431, 0.99,
            0.779, 0.855, 0.731, 0.665, 0.773, 0.647, 0.135, 0.44, 0.621, 0.329,
            0.718, 0.003, 0.927, 0.511, 0.515, 0.359, 0.744, 0.828, 0.31, 0.161,
            0.605, 0.539, 0.331, 0.077, 0.503, 0.668, 0.275, 0.72, 0.172, 0.035,
            0.88, 0.762, 0.646, 0.727, 0.83, 0.001, 0.085, 0.188, 0.583, 0.709,
            0.134, 0.683, 0.246, 0.214, 0.863, 0.109, 0.168, 0.539, 0.451, 0.303,
            0.064, 0.575, 0.547, 0.85, 0.75, 0.789, 0.681, 0.735], [0.648, 0.769,
            0.525, 0.716, 0.752, 0.199, 0.095, 0.222, 0.767, 0.029, 0.244, 0.527,
            0.496, 0.691, 0.487, 0.83, 0.546, 0.102, 0.845, 0.096, 0.744, 0.758,
            0.092, 0.289, 0.139, 0.005, 0.204, 0.245, 0.528, 0.607, 0.446, 0.029,
            0.686, 0.558, 0.705, 0.451, 0.87, 0.404, 0.824, 0.727, 0.058, 0.283,
            0.512, 0.682, 0.027, 0.026, 0.809, 0.669, 0.241, 0.103, 0.101, 0.225,
            0.989, 0.662, 0.917, 0.972, 0.93, 0.447, 0.318, 0.434, 0.437, 0.036,
            0.009, 0.96], [0.726, 0.418, 0.404, 0.244, 0.618, 0.356, 0.07, 0.842,
            0.137, 0.967, 0.465, 0.811, 0.027, 0.704, 0.935, 0.546, 0.92, 0.125,
            0.917, 0.089, 0.463, 0.929, 0.289, 0.721, 0.368, 0.837, 0.14, 0.431,
            0.495, 0.75, 0.484, 0.083, 0.431, 0.392, 0.177, 0.303, 0.013, 0.317,
            0.593, 0.047, 0.695, 0.185, 0.633, 0.825, 0.203, 0.619, 0.597, 0.152,
            0.899, 0.061, 0.512, 0.67, 0.82, 0.52, 0.743, 0.07, 0.99, 0.119,
            0.949, 0.284, 0.529, 0.65, 0.523, 0.059]]

        Supported metric type. Default is "L2" (L2, IP, HAMMING, TANIMOTO) [L2]: L2

        sqrt [False]: True

        Timeout(optional) []:
    """
    leftVectorMeta = {}
    if click.confirm("Import left operator vectors from existing collection?"):
        left_ids = click.prompt("The vectors' ids on the left of operator")
        left_collection = click.prompt(
            "The vectors' collection name on the left of operator")
        left_partition = click.prompt(
            "The vectors' partition name on the left of operator")
        left_field = click.prompt(
            "The vectors' field name on the left of operator")
        leftVectorMeta["vec_type"] = "import"
        leftVectorMeta["ids"] = left_ids
        leftVectorMeta["collection"] = left_collection
        leftVectorMeta["partition"] = left_partition
        leftVectorMeta["field"] = left_field
    else:
        left_type = click.prompt(
            "The vector's type",
            type=click.Choice(["float_vectors", "bin_vectors"]),
            default="float_vectors",
        )
        left_vectors = click.prompt("Left vectors")
        leftVectorMeta["vec_type"] = "raw"
        leftVectorMeta["type"] = left_type
        leftVectorMeta["vectors"] = left_vectors
    rightVectorMeta = {}
    if click.confirm(
            "Import right operator vectors from existing collection?"):
        right_ids = click.prompt("The vectors' ids on the right of operator")
        right_collection = click.prompt(
            "The vectors' collection name on the right of operator")
        right_partition = click.prompt(
            "The vectors' partition name on the right of operator")
        right_field = click.prompt(
            "The vectors' field name on the right of operator")
        rightVectorMeta["vec_type"] = "import"
        rightVectorMeta["ids"] = right_ids
        rightVectorMeta["collection"] = right_collection
        rightVectorMeta["partition"] = right_partition
        rightVectorMeta["field"] = right_field
    else:
        right_type = click.prompt(
            "The vector's type",
            type=click.Choice(["float_vectors", "bin_vectors"]),
            default="float_vectors",
        )
        right_vectors = click.prompt("Right vectors")
        rightVectorMeta["vec_type"] = "raw"
        rightVectorMeta["type"] = right_type
        rightVectorMeta["vectors"] = right_vectors
    metric_type = click.prompt(
        'Supported metric type. Default is "L2"',
        default="L2",
        type=click.Choice(MetricTypes),
    )
    sqrt = None
    if metric_type in ["L2"]:
        sqrt = click.prompt("sqrt", type=bool, default=False)
    dim = None
    if metric_type in ["HAMMING", "TANIMOTO"]:
        dim = click.prompt(
            "Set this value if dimension is not a multiple of 8, otherwise the dimension will be calculted by list length",
            type=int,
            default=None,
        )
    timeout = click.prompt("Timeout(optional)", default="")
    try:
        calcParams = validateCalcParams(leftVectorMeta, rightVectorMeta,
                                        metric_type, sqrt, dim, timeout)
        result = obj.calcDistance(
            calcParams["vectors_left"],
            calcParams["vectors_right"],
            calcParams["params"],
            calcParams["timeout"],
        )
    except ParameterException as pe:
        click.echo("Error!\n{}".format(str(pe)))
    except ConnectException as ce:
        click.echo("Error!\n{}".format(str(ce)))
    else:
        click.echo(
            "\n======\nReturn type:\n" +
            "Assume the vectors_left: L_1, L_2, L_3\n" +
            "Assume the vectors_right: R_a, R_b\n" +
            'Distance between L_n and R_m we called "D_n_m"\n' +
            "The returned distances are arranged like this:\n" +
            "   [[D_1_a, D_1_b],\n" + "   [D_2_a, D_2_b],\n" +
            "   [D_3_a, D_3_b]]\n" +
            '\nNote: if some vectors doesn\'t exist in collection, the returned distance is "-1.0"\n'
            + "======\n")
        click.echo("Result:\n")
        click.echo(result)


@calcUtils.command("mkts_from_hybridts")
@click.option(
    "-h",
    "--hybridts",
    "hybridts",
    help=
    "The original hybrid timestamp used to generate a new hybrid timestamp. Non-negative interger range from 0 to 18446744073709551615.",
    type=int,
)
@click.option(
    "-m",
    "--milliseconds",
    "milliseconds",
    help="Incremental time interval. The unit of time is milliseconds.",
    type=float,
    default=0.0,
)
# @click.option(
#     "-d",
#     "--delta",
#     "delta",
#     help="A duration expressing the difference between two date, time, or datetime instances to microsecond resolution.",
# )
@click.pass_obj
def hybridts2mkts(obj, hybridts, milliseconds):
    """Generate a hybrid timestamp based on an existing hybrid timestamp, timedelta and incremental time internval."""
    res = obj.mkts_from_hybridts(hybridts, milliseconds)
    click.echo(res)


@calcUtils.command("mkts_from_unixtime")
@click.option(
    "-e",
    "--epoch",
    "epoch",
    help=
    "The known Unix Epoch time used to generate a hybrid timestamp. The Unix Epoch time is the number of seconds that have elapsed since January 1, 1970 (midnight UTC/GMT).",
    type=float,
)
@click.option(
    "-m",
    "--milliseconds",
    "milliseconds",
    help="Incremental time interval. The unit of time is milliseconds.",
    type=float,
    default=0.0,
)
@click.pass_obj
def unixtime2mkts(obj, epoch, milliseconds):
    """Generate a hybrid timestamp based on Unix Epoch time, timedelta and incremental time internval."""
    res = obj.mkts_from_unixtime(epoch, milliseconds)
    click.echo(res)


@calcUtils.command("hybridts_to_unixtime")
@click.option(
    "-h",
    "--hybridts",
    "hybridts",
    help=
    "The known hybrid timestamp to convert to UNIX Epoch time. Non-negative interger range from 0 to 18446744073709551615.",
    type=int,
)
@click.pass_obj
def hybridts2unixtime(obj, hybridts):
    """Convert a hybrid timestamp to UNIX Epoch time ignoring the logic part."""
    res = obj.hybridts_to_unixtime(hybridts)
    click.echo(res)


@cli.command("load_balance")
@click.option(
    "-s",
    "--src-node-id",
    "src_node_id",
    help="The source query node id to balance.",
    type=int,
)
@click.option(
    "-d",
    "--dst-node-id",
    "dst_node_ids",
    help="[Multiple] - The destination query node ids to balance",
    type=int,
    multiple=True,
)
@click.option(
    "-ss",
    "--sealed-segment-ids",
    "sealed_segment_ids",
    help="[Multiple] - Sealed segment ids to balance.",
    type=int,
)
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help="[Optional] - The timeout for this method, unit: second",
    default=None,
    type=int,
)
@click.pass_obj
def loadBalance(obj, src_node_id, dst_node_ids, sealed_segment_ids, timeout):
    """Do load balancing operation from source query node to destination query node."""
    res = obj.loadBalance(src_node_id, dst_node_ids, sealed_segment_ids,
                          timeout)
    click.echo(res)


@cli.command("compact")
@click.option("-c",
              "--collection-name",
              "collectionName",
              help="Collection name")
@click.option(
    "-t",
    "--timeout",
    "timeout",
    help=
    "[Optional] - An optional duration of time in seconds to allow for the RPC. If timeout is not set, the client keeps waiting until the server responds or an error occurs.",
    default=None,
    type=float,
)
@click.pass_obj
def compactCollection(obj, collectionName, timeout):
    """Compact merge the small segments in a collection."""
    obj.compactCollection(collectionName, timeout)


@cli.command("exit")
def quitapp():
    """Exit the CLI."""
    global quitapp
    quitapp = True


quitapp = False  # global flag
comp = Completer()


def runCliPrompt():
    args = sys.argv
    if args and (args[-1] == "--version"):
        print(f"Milvus_CLI v{getPackageVersion()}")
        return
    try:
        print(WELCOME_MSG)
        while not quitapp:
            import readline

            readline.set_completer_delims(" \t\n;")
            readline.parse_and_bind("tab: complete")
            readline.set_completer(comp.complete)
            astr = input("milvus_cli > ")
            try:
                cli(astr.split())
            except SystemExit:
                # trap argparse error message
                # print('error', SystemExit)
                continue
            except ParameterException as pe:
                click.echo(message=f"{str(pe)}", err=True)
            except ConnectException as ce:
                click.echo(
                    message=
                    "Connect to milvus Error!\nPlease check your connection.",
                    err=True,
                )
            except Exception as e:
                click.echo(message=f"Error occurred!\n{str(e)}", err=True)
        print(EXIT_MSG)
    except (KeyboardInterrupt, EOFError):
        print(EXIT_MSG)
        sys.exit(0)


if __name__ == "__main__":
    runCliPrompt()



from tabulate import tabulate
import readline
import re
import os
from functools import reduce
from Types import DataTypeByNum
from Types import ParameterException, ConnectException
from time import time
from string import Template
from pymilvus import __version__


def getPackageVersion():
    import pkg_resources  # part of setuptools

    try:
        version = pkg_resources.require("milvus_cli")[0].version
    except Exception as e:
        raise ParameterException(
            "Could not get version under single executable file mode.")
    return version


def checkEmpty(x):
    return not not x


def getMilvusTimestamp(isSimilar=True):
    ts = time()
    if isSimilar:
        return int(ts) << 18


class PyOrm(object):
    host = "127.0.0.1"
    port = 19530
    alias = "default"

    def connect(self,
                alias=None,
                host=None,
                port=None,
                disconnect=False,
                secure=False,
                username=None,
                password=None):
        self.alias = alias
        self.host = host
        self.port = port
        trimUsername = None if username is None else username.strip()
        trimPwd = None if password is None else password.strip()

        from pymilvus import connections

        if disconnect:
            connections.disconnect(alias)
            return
        try:
            connections.connect(self.alias,
                                host=self.host,
                                port=self.port,
                                user=trimUsername,
                                password=trimPwd,
                                secure=secure)
        except Exception as e:
            raise ConnectException(f"Connect to Milvus error!{str(e)}")

    def checkConnection(self):
        from pymilvus import list_collections

        try:
            list_collections(timeout=10.0, using=self.alias)
        except Exception as e:
            raise ConnectException(f"Connect to Milvus error!{str(e)}")

    def showConnection(self, alias, showAll=False):
        from pymilvus import connections
        tempAlias = alias if alias else self.alias
        allConnections = connections.list_connections()
        if showAll:
            return tabulate(allConnections,
                            headers=["Alias", "Instance"],
                            tablefmt="pretty")
        aliasList = map(lambda x: x[0], allConnections)

        if tempAlias in aliasList:
            address, user = connections.get_connection_addr(tempAlias).values()
            # return """Host: {}\nPort: {}\nAlias: {}""".format(host, port, alias)
            return tabulate(
                [["Address", address], ["User", user], ["Alias", tempAlias]],
                tablefmt="pretty",
            )
        else:
            return "Connection not found!"

    def _list_collection_names(self, timeout=None):
        from pymilvus import utility

        return list(utility.list_collections(timeout, self.alias))

    def _list_partition_names(self, collectionName):
        target = self.getTargetCollection(collectionName)
        result = target.partitions
        return [i.name for i in result]

    def _list_field_names(self, collectionName, showVectorOnly=False):
        target = self.getTargetCollection(collectionName)
        result = target.schema.fields
        if showVectorOnly:
            return reduce(
                lambda x, y: x + [y.name]
                if y.dtype in [100, 101] else x, result, [])
        return [i.name for i in result]

    def _list_index(self, collectionName):
        target = self.getTargetCollection(collectionName)
        try:
            result = target.index()
        except Exception as e:
            return {}
        else:
            details = {
                "field_name": result.field_name,
                "index_type": result.params["index_type"],
                "metric_type": result.params["metric_type"],
                "params": result.params["params"],
            }
            # for p in result.params['params']:
            #     details[p] = result.params['params'][p]
            return details

    def listCollections(self, timeout=None):
        result = []
        collectionNames = self._list_collection_names(timeout)
        for name in collectionNames:
            result.append([name])

        return tabulate(
            result,
            headers=["Collection Name"],
            tablefmt="grid",
            showindex=True,
        )

    def flushCollectionByNumEntities(self, collectionName):
        col = self.getTargetCollection(collectionName)
        return col.num_entities

    def showCollectionLoadingProgress(self,
                                      collectionName,
                                      partition_names=None):
        from pymilvus import utility

        self.flushCollectionByNumEntities(collectionName)
        return utility.loading_progress(collectionName, partition_names,
                                        self.alias)

    def showIndexBuildingProgress(self, collectionName, index_name=""):
        from pymilvus import index_building_progress

        return index_building_progress(collectionName, index_name, self.alias)

    def getTargetCollection(self, collectionName):
        from pymilvus import Collection

        try:
            target = Collection(collectionName)
        except Exception as e:
            raise ParameterException("Collection error!\n")
        else:
            return target

    def getTargetPartition(self, collectionName, partitionName):
        try:
            targetCollection = self.getTargetCollection(collectionName)
            target = targetCollection.partition(partitionName)
        except Exception as e:
            raise ParameterException("Partition error!\n")
        else:
            return target

    def loadCollection(self, collectionName):
        target = self.getTargetCollection(collectionName)
        target.load()
        result = self.showCollectionLoadingProgress(collectionName)
        return tabulate(
            [[
                collectionName,
                result.get("num_loaded_entities"),
                result.get("num_total_entities"),
            ]],
            headers=["Collection Name", "Loaded", "Total"],
            tablefmt="grid",
        )

    def releaseCollection(self, collectionName):
        target = self.getTargetCollection(collectionName)
        target.release()
        result = self.showCollectionLoadingProgress(collectionName)
        return tabulate(
            [[
                collectionName,
                result.get("num_loaded_entities"),
                result.get("num_total_entities"),
            ]],
            headers=["Collection Name", "Loaded", "Total"],
            tablefmt="grid",
        )

    def releasePartition(self, collectionName, partitionName):
        targetPartition = self.getTargetPartition(collectionName,
                                                  partitionName)
        targetPartition.release()
        result = self.showCollectionLoadingProgress(collectionName,
                                                    [partitionName])
        return result

    def releasePartitions(self, collectionName, partitionNameList):
        result = []
        for name in partitionNameList:
            tmp = self.releasePartition(collectionName, name)
            result.append([
                name,
                tmp.get("num_loaded_entities"),
                tmp.get("num_total_entities")
            ])
        return tabulate(result,
                        headers=["Partition Name", "Loaded", "Total"],
                        tablefmt="grid")

    def loadPartition(self, collectionName, partitionName):
        targetPartition = self.getTargetPartition(collectionName,
                                                  partitionName)
        targetPartition.load()
        result = self.showCollectionLoadingProgress(collectionName,
                                                    [partitionName])
        return result

    def loadPartitions(self, collectionName, partitionNameList):
        result = []
        for name in partitionNameList:
            tmp = self.loadPartition(collectionName, name)
            result.append([
                name,
                tmp.get("num_loaded_entities"),
                tmp.get("num_total_entities")
            ])
        return tabulate(result,
                        headers=["Partition Name", "Loaded", "Total"],
                        tablefmt="grid")

    def listPartitions(self, collectionName):
        target = self.getTargetCollection(collectionName)
        result = target.partitions
        rows = list(map(lambda x: [x.name, x.description], result))
        return tabulate(
            rows,
            headers=["Partition Name", "Description"],
            tablefmt="grid",
            showindex=True,
        )

    def listIndexes(self, collectionName):
        target = self.getTargetCollection(collectionName)
        result = target.indexes

        rows = list(
            map(
                lambda x: [
                    x.field_name, x.index_name, x._index_params['index_type'],
                    x._index_params['metric_type'], x._index_params['params']
                ],
                result,
            ))
        return tabulate(
            rows,
            headers=[
                "Field Name", "Index Name", "Index Type", "Metric Type",
                "Params"
            ],
            tablefmt="grid",
            showindex=True,
        )

    def getCollectionDetails(self, collectionName="", collection=None):
        try:
            target = collection or self.getTargetCollection(collectionName)
        except Exception as e:
            return "Error!\nPlease check your input collection name."
        rows = []
        schema = target.schema
        partitions = target.partitions
        indexes = target.indexes
        fieldSchemaDetails = ""
        for fieldSchema in schema.fields:
            _name = f"{'*' if fieldSchema.is_primary else ''}{fieldSchema.name}"
            _type = DataTypeByNum[fieldSchema.dtype]
            _desc = fieldSchema.description
            _params = fieldSchema.params
            _dim = _params.get("dim")
            _params_desc = f"dim: {_dim}" if _dim else ""
            fieldSchemaDetails += f"\n - {_name} {_type} {_params_desc} {_desc}"
        schemaDetails = """Description: {}\n\nAuto ID: {}\n\nFields(* is the primary field):{}""".format(
            schema.description, schema.auto_id, fieldSchemaDetails)
        partitionDetails = "  - " + \
            "\n- ".join(map(lambda x: x.name, partitions))
        indexesDetails = "  - " + \
            "\n- ".join(map(lambda x: x.field_name, indexes))
        rows.append(["Name", target.name])
        rows.append(["Description", target.description])
        rows.append(["Is Empty", target.is_empty])
        rows.append(["Entities", target.num_entities])
        rows.append(["Primary Field", target.primary_field.name])
        rows.append(["Schema", schemaDetails])
        rows.append(["Partitions", partitionDetails])
        rows.append(["Indexes", indexesDetails])
        return tabulate(rows, tablefmt="grid")

    def getPartitionDetails(self, collection, partitionName=""):
        partition = collection.partition(partitionName)
        if not partition:
            return "No such partition!"
        rows = []
        rows.append(["Partition Name", partition.name])
        rows.append(["Description", partition.description])
        rows.append(["Is empty", partition.is_empty])
        rows.append(["Number of Entities", partition.num_entities])
        return tabulate(rows, tablefmt="grid")

    def getIndexDetails(self, collection, indexName):
        index = collection.index(index_name=indexName)
        if not index:
            return "No index!"
        rows = []
        rows.append(["Corresponding Collection", index.collection_name])
        rows.append(["Corresponding Field", index.field_name])
        rows.append(["Index Type", index.params["index_type"]])
        rows.append(["Metric Type", index.params["metric_type"]])
        params = index.params["params"]
        paramsDetails = "\n- ".join(
            map(lambda k: f"{k[0]}: {k[1]}", params.items()))
        rows.append(["Params", paramsDetails])
        return tabulate(rows, tablefmt="grid")

    def createCollection(self, collectionName, primaryField, autoId,
                         description, fields):
        from pymilvus import Collection, DataType, FieldSchema, CollectionSchema

        fieldList = []
        for field in fields:
            [fieldName, fieldType, fieldData] = field.split(":")
            isVector = False
            if fieldType in ["BINARY_VECTOR", "FLOAT_VECTOR"]:
                fieldList.append(
                    FieldSchema(name=fieldName,
                                dtype=DataType[fieldType],
                                dim=int(fieldData)))
            else:
                fieldList.append(
                    FieldSchema(name=fieldName,
                                dtype=DataType[fieldType],
                                description=fieldData))
        schema = CollectionSchema(
            fields=fieldList,
            primary_field=primaryField,
            auto_id=autoId,
            description=description,
        )
        collection = Collection(name=collectionName, schema=schema)
        return self.getCollectionDetails(collection=collection)

    def createPartition(self, collectionName, description, partitionName):
        collection = self.getTargetCollection(collectionName)
        collection.create_partition(partitionName, description=description)
        return self.getPartitionDetails(collection, partitionName)

    def createIndex(self, collectionName, fieldName, indexName, indexType,
                    metricType, params, timeout):
        collection = self.getTargetCollection(collectionName)
        indexParams = {}
        for param in params:
            paramList = param.split(":")
            [paramName, paramValue] = paramList
            indexParams[paramName] = int(paramValue)
        index = {
            "index_type": indexType,
            "params": indexParams,
            "metric_type": metricType,
        }
        collection.create_index(field_name=fieldName,
                                index_params=index,
                                index_name=indexName,
                                timeout=timeout)

        return self.getIndexDetails(collection, indexName)

    def isCollectionExist(self, collectionName):
        from pymilvus import has_collection

        return has_collection(collectionName, using=self.alias)

    def isPartitionExist(self, collection, partitionName):
        return collection.has_partition(partitionName)

    def isIndexExist(self, collection, indexName):
        return collection.has_index(index_name=indexName)

    def dropCollection(self, collectionName, timeout):
        collection = self.getTargetCollection(collectionName)
        collection.drop(timeout=timeout)
        return self.isCollectionExist(collectionName)

    def dropPartition(self, collectionName, partitionName, timeout):
        collection = self.getTargetCollection(collectionName)
        collection.drop_partition(partitionName, timeout=timeout)
        return self.isPartitionExist(collection, partitionName)

    def dropIndex(self, collectionName, indexName, timeout):
        collection = self.getTargetCollection(collectionName)
        collection.drop_index(index_name=indexName, timeout=timeout)
        return self.listIndexes(collectionName)

    def search(self, collectionName, searchParameters, prettierFormat=True):
        collection = self.getTargetCollection(collectionName)
        collection.load()
        res = collection.search(**searchParameters)
        if not prettierFormat:
            return res
        # hits = res[0]
        results = []
        for hits in res:
            results += [
                tabulate(
                    map(lambda x: [x.id, x.distance, x.score], hits),
                    headers=["Index", "ID", "Distance", "Score"],
                    tablefmt="grid",
                    showindex=True,
                )
            ]
        # return tabulate(map(lambda x: [x.id, x.distance], hits), headers=['Index', 'ID', 'Distance'], tablefmt='grid', showindex=True)
        return results

    def query(self, collectionName, queryParameters):
        collection = self.getTargetCollection(collectionName)
        print(queryParameters)
        res = collection.query(**queryParameters)
        # return f"- Query results: {res}"
        if not len(res):
            return f"- Query results: {res}"
        headers = [i for i in res[0]]
        return tabulate(
            [[_[i] for i in _] for _ in res],
            headers=headers,
            tablefmt="grid",
            showindex=True,
        )

    def insert(self, collectionName, data, partitionName=None, timeout=None):
        collection = self.getTargetCollection(collectionName)
        result = collection.insert(data,
                                   partition_name=partitionName,
                                   timeout=timeout)
        entitiesNum = collection.num_entities
        return [result, entitiesNum]

    def importData(self,
                   collectionName,
                   data,
                   partitionName=None,
                   timeout=None):
        [result, entitiesNum] = self.insert(collectionName, data,
                                            partitionName, timeout)
        insert_count = result.insert_count
        timestamp = result.timestamp
        prettierResult = []
        prettierResult.append(["Total insert entities: ", insert_count])
        prettierResult.append(["Total collection entities: ", entitiesNum])
        prettierResult.append(["Milvus timestamp: ", timestamp])
        return tabulate(prettierResult)

    def calcDistance(self,
                     vectors_left,
                     vectors_right,
                     params=None,
                     timeout=None):
        from pymilvus import utility

        result = utility.calc_distance(vectors_left,
                                       vectors_right,
                                       params,
                                       timeout,
                                       using=self.alias)
        return result

    def deleteEntities(self,
                       expr,
                       collectionName,
                       partition_name=None,
                       timeout=None):
        collection = self.getTargetCollection(collectionName)
        result = collection.delete(expr,
                                   partition_name=partition_name,
                                   timeout=timeout)
        return result

    def getQuerySegmentInfo(self,
                            collectionName,
                            timeout=None,
                            prettierFormat=False):
        from pymilvus import utility

        result = utility.get_query_segment_info(collectionName,
                                                timeout=timeout,
                                                using=self.alias)
        if not prettierFormat or not result:
            return result
        firstChild = result[0]
        headers = [
            "segmentID", "collectionID", "partitionID", "mem_size", "num_rows"
        ]
        return tabulate(
            [[getattr(_, i) for i in headers] for _ in result],
            headers=headers,
            showindex=True,
        )

    def createCollectionAlias(self,
                              collectionName,
                              collectionAliasName="",
                              timeout=None):
        from pymilvus import utility

        collection = self.getTargetCollection(collectionName)
        result = utility.create_alias(collection.name,
                                      collectionAliasName,
                                      timeout=timeout)
        return result

    def dropCollectionAlias(self, collectionAliasName="", timeout=None):
        from pymilvus import utility

        result = utility.drop_alias(collectionAliasName, timeout=timeout)
        return result

    def alterCollectionAlias(self,
                             collectionName,
                             collectionAliasName="",
                             timeout=None):
        from pymilvus import utility

        collection = self.getTargetCollection(collectionName)
        result = utility.alter_alias(collection.name,
                                     collectionAliasName,
                                     timeout=timeout)
        return result

    def createCollectionAliasList(self,
                                  collectionName,
                                  aliasList=[],
                                  timeout=None):
        from pymilvus import utility

        collection = self.getTargetCollection(collectionName)
        result = []
        for aliasItem in aliasList:
            aliasResult = utility.create_alias(collection.name,
                                               aliasItem,
                                               timeout=timeout)
            result.append(aliasResult)
        return result

    def alterCollectionAliasList(self,
                                 collectionName,
                                 aliasList=[],
                                 timeout=None):
        from pymilvus import utility

        collection = self.getTargetCollection(collectionName)
        result = []
        for aliasItem in aliasList:
            aliasResult = utility.alter_alias(collection.name,
                                              aliasItem,
                                              timeout=timeout)
            result.append(aliasResult)
        return result

    def listCollectionAlias(self, collectionName, timeout=None):
        from pymilvus import utility

        collection = self.getTargetCollection(collectionName)
        result = utility.list_aliases(collection.name, timeout=timeout)
        print("alias==>", result)
        return result

    def loadBalance(self,
                    src_node_id,
                    dst_node_ids,
                    sealed_segment_ids,
                    timeout=None):
        from pymilvus import utility

        res = utility.load_balance(src_node_id,
                                   dst_node_ids,
                                   sealed_segment_ids,
                                   timeout=timeout)
        return res

    def mkts_from_hybridts(self, hybridts, milliseconds=0.0):
        from pymilvus import utility

        ts_new = utility.mkts_from_hybridts(hybridts, milliseconds)
        return ts_new

    def mkts_from_unixtime(self, epoch, milliseconds=0.0):
        from pymilvus import utility

        ts_new = utility.mkts_from_unixtime(epoch, milliseconds)
        return ts_new

    def hybridts_to_unixtime(self, hybridts):
        from pymilvus import utility

        ts_new = utility.hybridts_to_unixtime(hybridts)
        return ts_new

    # pymilvus.utility.mkts_from_datetime(d_time, milliseconds=0.0, delta=None)
    # pymilvus.utility.hybridts_to_datetime(hybridts, tz=None)

    def compactCollection(self, collectionName, timeout=None):
        collection = self.getTargetCollection(collectionName)
        collection.compact(timeout=timeout)

    def getCompactCollectionState(self, collectionName, timeout=None):
        collection = self.getTargetCollection(collectionName)
        return collection.get_compaction_state(timeout=timeout)

    def getCompactCollectionPlans(self, collectionName, timeout=None):
        collection = self.getTargetCollection(collectionName)
        return collection.get_compaction_plans(timeout=timeout)

    def listCredUsers(self):
        from pymilvus import utility
        users = utility.list_usernames(self.alias)
        return users

    def createCredUser(self, username=None, password=None):
        from pymilvus import utility
        utility.create_user(username, password, self.alias)
        return self.listCredUsers()

    def deleteCredUser(self, username=None):
        from pymilvus import utility
        utility.delete_user(username, self.alias)
        return self.listCredUsers()


class Completer(object):
    # COMMANDS = ['clear', 'connect', 'create', 'delete', 'describe', 'exit',
    #         'list', 'load', 'query', 'release', 'search', 'show', 'version' ]
    RE_SPACE = re.compile(".*\s+$", re.M)
    CMDS_DICT = {
        "calc": [
            "distance",
            "mkts_from_hybridts",
            "mkts_from_unixtime",
            "hybridts_to_unixtime",
        ],
        "clear": [],
        "compact": [],
        "connect": [],
        "create": ["alias", "collection", "partition", "index", "user"],
        "delete":
        ["alias", "collection", "entities", "partition", "index", "user"],
        "describe": ["collection", "partition", "index"],
        "exit": [],
        "help": [],
        "import": [],
        "list": ["collections", "partitions", "indexes", "users"],
        "load_balance": [],
        "load": [],
        "query": [],
        "release": [],
        "search": [],
        "show": [
            "connection",
            "index_progress",
            "loading_progress",
            "query_segment",
            "compaction_state",
            "compaction_plans",
        ],
        "version": [],
    }

    def __init__(self) -> None:
        super().__init__()
        self.COMMANDS = list(self.CMDS_DICT.keys())
        self.createCompleteFuncs(self.CMDS_DICT)

    def createCompleteFuncs(self, cmdDict):
        for cmd in cmdDict:
            sub_cmds = cmdDict[cmd]
            complete_example = self.makeComplete(cmd, sub_cmds)
            setattr(self, "complete_%s" % cmd, complete_example)

    def makeComplete(self, cmd, sub_cmds):

        def f_complete(args):
            f"Completions for the {cmd} command."
            if not args:
                return self._complete_path(".")
            if len(args) <= 1 and not cmd == "import":
                return self._complete_2nd_level(sub_cmds, args[-1])
            return self._complete_path(args[-1])

        return f_complete

    def _listdir(self, root):
        "List directory 'root' appending the path separator to subdirs."
        res = []
        for name in os.listdir(root):
            path = os.path.join(root, name)
            if os.path.isdir(path):
                name += os.sep
            res.append(name)
        return res

    def _complete_path(self, path=None):
        "Perform completion of filesystem path."
        if not path:
            return self._listdir(".")
        dirname, rest = os.path.split(path)
        tmp = dirname if dirname else "."
        res = [
            os.path.join(dirname, p) for p in self._listdir(tmp)
            if p.startswith(rest)
        ]
        # more than one match, or single match which does not exist (typo)
        if len(res) > 1 or not os.path.exists(path):
            return res
        # resolved to a single directory, so return list of files below it
        if os.path.isdir(path):
            return [os.path.join(path, p) for p in self._listdir(path)]
        # exact file match terminates this completion
        return [path + " "]

    def _complete_2nd_level(self, SUB_COMMANDS=[], cmd=None):
        if not cmd:
            return [c + " " for c in SUB_COMMANDS]
        res = [c for c in SUB_COMMANDS if c.startswith(cmd)]
        if len(res) > 1 or not (cmd in SUB_COMMANDS):
            return res
        return [cmd + " "]

    # def complete_create(self, args):
    #     "Completions for the 'create' command."
    #     if not args:
    #         return self._complete_path('.')
    #     sub_cmds = ['collection', 'partition', 'index']
    #     if len(args) <= 1:
    #         return self._complete_2nd_level(sub_cmds, args[-1])
    #     return self._complete_path(args[-1])

    def complete(self, text, state):
        "Generic readline completion entry point."
        buffer = readline.get_line_buffer()
        line = readline.get_line_buffer().split()
        # show all commands
        if not line:
            return [c + " " for c in self.COMMANDS][state]
        # account for last argument ending in a space
        if self.RE_SPACE.match(buffer):
            line.append("")
        # resolve command to the implementation function
        cmd = line[0].strip()
        if cmd in self.COMMANDS:
            impl = getattr(self, "complete_%s" % cmd)
            args = line[1:]
            if args:
                return (impl(args) + [None])[state]
            return [cmd + " "][state]
        results = [c + " "
                   for c in self.COMMANDS if c.startswith(cmd)] + [None]
        return results[state]


msgTemp = Template("""


  __  __ _ _                    ____ _     ___
 |  \/  (_) |_   ___   _ ___   / ___| |   |_ _|
 | |\/| | | \ \ / / | | / __| | |   | |    | |
 | |  | | | |\ V /| |_| \__ \ | |___| |___ | |
 |_|  |_|_|_| \_/  \__,_|___/  \____|_____|___|

Milvus cli version: ${cli}
Pymilvus version: ${py}

Learn more: https://github.com/zilliztech/milvus_cli.

""")

WELCOME_MSG = msgTemp.safe_substitute(cli=getPackageVersion(), py=__version__)

EXIT_MSG = "\n\nThanks for using.\nWe hope your feedback: https://github.com/zilliztech/milvus_cli/issues/new.\n\n"



import os, csv, random

length = 50

vectors = [[random.random() for _ in range(128)] for _ in range(length)]

with open(f"{os.path.split(os.path.realpath(__file__))[0]}/vectors.csv", "w+") as f:
    writer = csv.writer(f)
    writer.writerow(["vector", "color", "brand"])
    for idx, vector in enumerate(vectors):
        writer.writerow([vector, idx + 2000, idx])



from Types import ParameterException
import os


def readCsvFile(path="", withCol=True):
    import re

    pattern = re.compile(".+\..+\/.*")
    isUrl = pattern.match(path)
    if isUrl:
        return readCsvFileFromUrl(path, withCol)
    return readCsvFileFromLocal(path, withCol)


def readCsvFileFromLocal(path="", withCol=True):
    from csv import reader
    from json import JSONDecodeError
    import click

    click.echo("Reading file from local path.")
    if not path or not path[-4:] == ".csv":
        raise ParameterException("Path is empty or target file is not .csv")
    fileSize = os.stat(path).st_size
    if fileSize >= 512000000:
        raise ParameterException(
            "File is too large! Only allow csv files less than 512MB."
        )
    try:
        result = {"columns": [], "data": []}
        with click.open_file(path, "r") as csv_file:
            click.echo(f"Opening csv file({fileSize} bytes)...")
            csv_reader = reader(csv_file, delimiter=",")
            handleCsvFile(result, csv_reader, withCol)
    except FileNotFoundError as fe:
        raise ParameterException(f"FileNotFoundError {str(fe)}")
    except UnicodeDecodeError as ue:
        raise ParameterException(f"UnicodeDecodeError {str(ue)}")
    except JSONDecodeError as je:
        raise ParameterException(f"JSONDecodeError {str(je)}")
    else:
        return result


def readCsvFileFromUrl(url="", withCol=True):
    import requests
    from csv import reader
    from json import JSONDecodeError
    import click

    click.echo("Reading file from remote URL.")
    try:
        result = {"columns": [], "data": []}
        with requests.Session() as s:
            download = s.get(url)
            decoded_content = download.content.decode("utf-8")
            csv_reader = reader(decoded_content.splitlines(), delimiter=",")
            handleCsvFile(result, csv_reader, withCol)
    except FileNotFoundError as fe:
        raise ParameterException(f"FileNotFoundError {str(fe)}")
    except UnicodeDecodeError as ue:
        raise ParameterException(f"UnicodeDecodeError {str(ue)}")
    except JSONDecodeError as je:
        raise ParameterException(f"JSONDecodeError {str(je)}")
    else:
        return result


def handleCsvFile(result, csv_reader, withCol):
    import click

    # For progressbar, transform it to list.
    rows = list(csv_reader)
    line_count = 0
    with click.progressbar(rows, label="Reading csv rows...", show_percent=True) as bar:
        # for row in csv_reader:
        for row in bar:
            if withCol and line_count == 0:
                result["columns"] = row
                line_count += 1
            else:
                formatRowForData(row, result["data"])
                line_count += 1
    click.echo(f"""Column names are {result['columns']}""")
    click.echo(f"Processed {line_count} lines.")


# For readCsvFile formatting data.
def formatRowForData(row=[], data=[]):
    from json import loads

    # init data with empty list
    if not data:
        for _in in range(len(row)):
            data.append([])
    for idx, val in enumerate(row):
        formattedVal = loads(val)
        data[idx].append(formattedVal)


def writeCsvFile(path, rows, headers=[]):
    if not path:
        raise ParameterException(f"Path should not be empty")
    from csv import writer
    import click

    try:
        with click.open_file(path, "w+") as csv_file:
            csv_writer = writer(csv_file, delimiter=",")
            if headers:
                csv_writer.writerow(headers)
            line_count = 0
            with click.progressbar(
                rows, label="Writing csv rows...", show_percent=True
            ) as bar:
                for row in bar:
                    csv_writer.writerow(row)
                    line_count += 1
            click.echo(f"Processed {line_count} lines.")
    except Exception as e:
        raise ParameterException(f"Export csv file error! {str(e)}")



from functools import reduce


class ParameterException(Exception):
    "Custom Exception for parameters checking."

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class ConnectException(Exception):
    "Custom Exception for milvus connection."

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


FiledDataTypes = [
    "BOOL",
    "INT8",
    "INT16",
    "INT32",
    "INT64",
    "FLOAT",
    "DOUBLE",
    "STRING",
    "BINARY_VECTOR",
    "FLOAT_VECTOR",
]

IndexTypes = [
    "FLAT",
    "IVF_FLAT",
    "IVF_SQ8",
    "IVF_PQ",
    "RNSG",
    "HNSW",
    # "NSG",
    "ANNOY",
    # "RHNSW_FLAT",
    # "RHNSW_PQ",
    # "RHNSW_SQ",
    # "BIN_FLAT",
    # "BIN_IVF_FLAT"
]

IndexParams = [
    "nlist",
    "m",
    "nbits",
    "M",
    "efConstruction",
    "n_trees",
    "PQM",
]

IndexTypesMap = {
    "FLAT": {
        "index_building_parameters": [],
        "search_parameters": ["metric_type"],
    },
    "IVF_FLAT": {
        "index_building_parameters": ["nlist"],
        "search_parameters": ["nprobe"],
    },
    "IVF_SQ8": {
        "index_building_parameters": ["nlist"],
        "search_parameters": ["nprobe"],
    },
    "IVF_PQ": {
        "index_building_parameters": ["nlist", "m", "nbits"],
        "search_parameters": ["nprobe"],
    },
    "RNSG": {
        "index_building_parameters": [
            "out_degree",
            "candidate_pool_size",
            "search_length",
            "knng",
        ],
        "search_parameters": ["search_length"],
    },
    "HNSW": {
        "index_building_parameters": ["M", "efConstruction"],
        "search_parameters": ["ef"],
    },
    "ANNOY": {
        "index_building_parameters": ["n_trees"],
        "search_parameters": ["search_k"],
    },
}

DupSearchParams = reduce(
    lambda x, y: x + IndexTypesMap[y]["search_parameters"], IndexTypesMap.keys(), []
)
SearchParams = list(dict.fromkeys(DupSearchParams))

MetricTypes = ["L2", "IP", "HAMMING", "TANIMOTO"]

DataTypeByNum = {
    0: "NONE",
    1: "BOOL",
    2: "INT8",
    3: "INT16",
    4: "INT32",
    5: "INT64",
    10: "FLOAT",
    11: "DOUBLE",
    20: "STRING",
    21: "VARCHAR",
    100: "BINARY_VECTOR",
    101: "FLOAT_VECTOR",
    999: "UNKNOWN",
}

Operators = ["<", "<=", ">", ">=", "==", "!=", "in"]



from Types import ParameterException
from Types import (
    FiledDataTypes,
    IndexTypes,
    IndexTypesMap,
    SearchParams,
    MetricTypes,
    Operators,
)
from Fs import readCsvFile


def validateParamsByCustomFunc(customFunc, errMsg, *params):
    try:
        customFunc(*params)
    except Exception as e:
        raise ParameterException(f"{errMsg}")


def validateCollectionParameter(collectionName, primaryField, fields):
    if not collectionName:
        raise ParameterException("Missing collection name.")
    if not primaryField:
        raise ParameterException("Missing primary field.")
    if not fields:
        raise ParameterException("Missing fields.")
    fieldNames = []
    for field in fields:
        fieldList = field.split(":")
        if not (len(fieldList) == 3):
            raise ParameterException(
                'Field should contain three parameters concatenated by ":".'
            )
        [fieldName, fieldType, fieldData] = fieldList
        fieldNames.append(fieldName)
        if fieldType not in FiledDataTypes:
            raise ParameterException(
                "Invalid field data type, should be one of {}".format(
                    str(FiledDataTypes)
                )
            )
        if fieldType in ["BINARY_VECTOR", "FLOAT_VECTOR"]:
            try:
                int(fieldData)
            except ValueError as e:
                raise ParameterException("""Vector's dim should be int.""")
    # Dedup field name.
    newNames = list(set(fieldNames))
    if not (len(newNames) == len(fieldNames)):
        raise ParameterException("Field names are duplicated.")
    if primaryField not in fieldNames:
        raise ParameterException(
            """Primary field name doesn't exist in input fields."""
        )


def validateIndexParameter(indexType, metricType, params):
    if indexType not in IndexTypes:
        raise ParameterException(
            "Invalid index type, should be one of {}".format(str(IndexTypes))
        )
    if metricType not in MetricTypes:
        raise ParameterException(
            "Invalid index metric type, should be one of {}".format(str(MetricTypes))
        )
    # if not params:
    #     raise ParameterException('Missing params')
    paramNames = []
    buildingParameters = IndexTypesMap[indexType]["index_building_parameters"]
    for param in params:
        paramList = param.split(":")
        if not (len(paramList) == 2):
            raise ParameterException(
                'Params should contain two parameters concatenated by ":".'
            )
        [paramName, paramValue] = paramList
        paramNames.append(paramName)
        if paramName not in buildingParameters:
            raise ParameterException(
                "Invalid index param, should be one of {}".format(
                    str(buildingParameters)
                )
            )
        try:
            int(paramValue)
        except ValueError as e:
            raise ParameterException("""Index param's value should be int.""")
    # Dedup field name.
    newNames = list(set(paramNames))
    if not (len(newNames) == len(paramNames)):
        raise ParameterException("Index params are duplicated.")


def validateSearchParams(
    data,
    annsField,
    metricType,
    params,
    limit,
    expr,
    partitionNames,
    timeout,
    roundDecimal,
    hasIndex=True,
    guarantee_timestamp=None,
    travel_timestamp=None,
):
    import json

    result = {}
    # Validate data
    try:
        if ".csv" in data:
            csvData = readCsvFile(data, withCol=False)
            result["data"] = csvData["data"][0]
        else:
            result["data"] = json.loads(data.replace("'", "").replace('"', ""))
    except Exception as e:
        raise ParameterException(
            'Format(list[list[float]]) "Data" error! {}'.format(str(e))
        )
    # Validate annsField
    if not annsField:
        raise ParameterException("annsField is empty!")
    result["anns_field"] = annsField
    if hasIndex:
        # Validate metricType
        if metricType not in MetricTypes:
            raise ParameterException(
                "Invalid index metric type, should be one of {}".format(
                    str(MetricTypes)
                )
            )
        # Validate params
        paramDict = {}
        if type(params) == str:
            paramsList = params.replace(" ", "").split(",")
        else:
            paramsList = params
        for param in paramsList:
            if not param:
                continue
            paramList = param.split(":")
            if not (len(paramList) == 2):
                raise ParameterException(
                    'Params should contain two parameters concatenated by ":".'
                )
            [paramName, paramValue] = paramList
            if paramName not in SearchParams:
                raise ParameterException(
                    "Invalid search parameter, should be one of {}".format(
                        str(SearchParams)
                    )
                )
            try:
                paramDict[paramName] = int(paramValue)
            except ValueError as e:
                raise ParameterException("""Search parameter's value should be int.""")
        result["param"] = {"metric_type": metricType}
        if paramDict.keys():
            result["param"]["params"] = paramDict
    else:
        result["param"] = {}
    #  Validate limit
    try:
        result["limit"] = int(limit)
    except Exception as e:
        raise ParameterException('Format(int) "limit" error! {}'.format(str(e)))
    # Validate expr
    result["expr"] = expr if expr else None
    # Validate partitionNames
    if partitionNames:
        try:
            result["partition_names"] = partitionNames.replace(" ", "").split(",")
        except Exception as e:
            raise ParameterException(
                'Format(list[str]) "partitionNames" error! {}'.format(str(e))
            )
    # Validate timeout
    if timeout:
        result["timeout"] = float(timeout)
    if roundDecimal:
        result["round_decimal"] = int(roundDecimal)
    #  Validate guarantee_timestamp and travel_timestamp
    if guarantee_timestamp:
        try:
            result["guarantee_timestamp"] = int(guarantee_timestamp)
        except Exception as e:
            raise ParameterException(
                'Format(int) "guarantee_timestamp" error! {}'.format(str(e))
            )
    if travel_timestamp:
        try:
            result["travel_timestamp"] = int(travel_timestamp)
        except Exception as e:
            raise ParameterException(
                'Format(int) "travel_timestamp" error! {}'.format(str(e))
            )
    return result


def validateQueryParams(
    expr,
    partitionNames,
    outputFields,
    timeout,
    guarantee_timestamp,
    graceful_time,
    travel_timestamp,
):
    result = {}
    if not expr:
        raise ParameterException("expr is empty!")
    # if ' in ' not in expr:
    if not any(map(lambda x: x in expr, Operators)):
        raise ParameterException(
            f'The query expression only accepts "<field_name> <oprator in {Operators}> [<value>, ...]"!'
        )
    result["expr"] = expr
    if not outputFields:
        result["output_fields"] = None
    else:
        nameList = outputFields.replace(" ", "").split(",")
        result["output_fields"] = nameList
    if not partitionNames:
        result["partition_names"] = None
    else:
        nameList = partitionNames.replace(" ", "").split(",")
        result["partition_names"] = nameList
    result["timeout"] = float(timeout) if timeout else None
    if guarantee_timestamp:
        result["guarantee_timestamp"] = guarantee_timestamp
    if graceful_time and (graceful_time != 5):
        result["graceful_time"] = graceful_time
    if travel_timestamp:
        result["travel_timestamp"] = travel_timestamp
    return result


def validateCalcParams(
    leftVectorMeta, rightVectorMeta, metric_type, sqrt, dim, timeout
):
    result = {"params": {}}
    vectors_left = validateVectorMeta(leftVectorMeta)
    result["vectors_left"] = vectors_left
    vectors_right = validateVectorMeta(rightVectorMeta)
    result["vectors_right"] = vectors_right
    params = result["params"]
    params["metric_type"] = metric_type
    if metric_type not in MetricTypes:
        raise ParameterException(
            "metric_type should be one of {}".format(str(MetricTypes))
        )
    if metric_type == "L2":
        params["sqrt"] = sqrt
    elif metric_type in ["HAMMING", "TANIMOTO"]:
        params["dim"] = dim
    if timeout:
        result["timeout"] = float(timeout)
    else:
        result["timeout"] = None
    return result


def validateVectorMeta(vectorMeta):
    import json

    vec_type = vectorMeta["vec_type"]
    result = {}
    if vec_type == "import":
        ids_str = vectorMeta["ids"]
        try:
            ids = json.loads(ids_str.replace("'", "").replace('"', ""))
        except Exception as e:
            raise ParameterException('Format(list[int]) "ids" error! {}'.format(str(e)))
        collection_str = vectorMeta["collection"]
        partition_str = vectorMeta["partition"]
        field_str = vectorMeta["field"]
        if not collection_str or not partition_str or not field_str:
            raise ParameterException(f"Collection/Partition/Field should not be empty!")
        result["ids"] = ids
        result["collection"] = collection_str
        result["partition"] = partition_str
        result["field"] = field_str
    if vec_type == "raw":
        vectors_str = vectorMeta["vectors"]
        if vectorMeta["type"] == "float_vectors":
            try:
                vectors = json.loads(vectors_str.replace("'", "").replace('"', ""))
            except Exception as e:
                raise ParameterException(
                    'Format(list[float]) "ids" error! {}'.format(str(e))
                )
            else:
                result[vectorMeta["type"]] = vectors
        elif vectorMeta["type"] == "bin_vectors":
            # """[b'\x94', b'N', b'\xca']"""
            noWhiteSpace = vectors_str.strip()
            noSquareBrackets = noWhiteSpace[1:-1]
            # "b'\x94', b'N', b'Ê'"
            strList = noSquareBrackets.split(", ")
            binMap = map(lambda x: x[2:-1].encode("unicode_escape"), strList)
            result[vectorMeta["type"]] = list(binMap)
    return result



from setuptools import setup, find_namespace_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="milvus_cli",
    version="v0.3.1",
    author="Milvus Team",
    author_email="milvus-team@zilliz.com",
    url="https://github.com/zilliztech/milvus_cli",
    description="CLI for Milvus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    packages=find_namespace_packages(),
    include_package_data=True,
    install_requires=[
        "Click==8.0.1",
        "pymilvus==2.2.1",
        "tabulate==0.8.9",
        "requests==2.26.0",
    ],
    entry_points={
        "console_scripts": [
            "milvus_cli = milvus_cli.scripts.milvus_cli:runCliPrompt",
        ],
    },
    python_requires=">=3.8",
)
